# -*- coding: utf-8 -*-
"""Plugin to test the target server for CVE-2014-0224.
"""

from xml.etree.ElementTree import Element

from sslyze.plugins import plugin_base
from nassl import TLSV1, TLSV1_1, TLSV1_2, SSLV3
import socket, struct, time, random

from sslyze.plugins.plugin_base import PluginResult


class OpenSslCcsInjectionPlugin(plugin_base.PluginBase):

    interface = plugin_base.PluginInterface("OpenSslCcsInjectionPlugin",  "")
    interface.add_command(
        command="openssl_ccs",
        help="Tests the server(s) for the OpenSSL CCS injection vulnerability (experimental)."
    )

    def srecv(self):
        r = self._sock.recv(4096)
        self._inbuffer += r
        return r != ''

    def process_task(self, server_info, plugin_command, options_dict=None):

        ssl_connection = server_info.get_preconfigured_ssl_connection()
        self._ssl_version = server_info.highest_ssl_version_supported
        is_vulnerable = False
        self._inbuffer = ""
        ssl_connection.do_pre_handshake()

        # H4ck to directly send the CCS payload
        self._sock = ssl_connection._sock

        # Send hello and wait for server hello & cert
        serverhello, servercert = False, False
        self._sock.send(self.make_hello())
        while not serverhello: #  or not servercert
            try:
                if not self.srecv(): break
            except:
                break
            rs = self.parse_records()
            for record in rs:
                if record['type'] == 22:
                    for p in record['proto']:
                        if p['type'] == 2:
                            serverhello= True
                        if p['type'] == 11:
                            servercert= True

        # Send the CCS
        if serverhello: # and servercert:
            is_vulnerable, stop = True, False
            self._sock.send(self.make_ccs())
            while not stop:
                try:
                    if not self.srecv(): break
                except socket.timeout:
                    break
                except:
                    is_vulnerable = False
                    stop = True

                rs = self.parse_records()
                for record in rs:
                    if record['type'] == 21:
                        for p in record['proto']:
                            if p['level'] == 2 or (p['level'] == 1 and p['desc'] == 0):
                                is_vulnerable = False
                                stop = True

            # If we receive no alert message check whether it is really is_vulnerable
            if is_vulnerable:
                self._sock.send('\x15' + self.ssl_tokens[self._ssl_version] + '\x00\x02\x01\x00')

                try:
                    if not self.srecv():
                        is_vulnerable = False
                except:
                    is_vulnerable = False

        self._sock.close()
        return OpenSslCcsInjectionResult(server_info, plugin_command, options_dict, is_vulnerable)


    ssl_tokens = {
        SSLV3   : "\x03\x00",
        TLSV1   : "\x03\x01",
        TLSV1_1 : "\x03\x02",
        TLSV1_2 : "\x03\x03",
    }

    ssl3_cipher = [
        '\x00\x00', '\x00\x01', '\x00\x02', '\x00\x03',
        '\x00\x04', '\x00\x05', '\x00\x06', '\x00\x07',
        '\x00\x08', '\x00\x09', '\x00\x0a', '\x00\x0b',
        '\x00\x0c', '\x00\x0d', '\x00\x0e', '\x00\x0f',
        '\x00\x10', '\x00\x11', '\x00\x12', '\x00\x13',
        '\x00\x14', '\x00\x15', '\x00\x16', '\x00\x17',
        '\x00\x18', '\x00\x19', '\x00\x1a', '\x00\x1b',
        '\x00\x1c', '\x00\x1d', '\x00\x1e',
        '\x00\x1F', '\x00\x20', '\x00\x21', '\x00\x22',
        '\x00\x23', '\x00\x24', '\x00\x25', '\x00\x26',
        '\x00\x27', '\x00\x28', '\x00\x29', '\x00\x2A',
        '\x00\x2B', '\x00\x2C', '\x00\x2D', '\x00\x2E',
        '\x00\x2F', '\x00\x30', '\x00\x31', '\x00\x32',
        '\x00\x33', '\x00\x34', '\x00\x35', '\x00\x36',
        '\x00\x37', '\x00\x38', '\x00\x39', '\x00\x3A',
        '\x00\x3B', '\x00\x3C', '\x00\x3D', '\x00\x3E',
        '\x00\x3F', '\x00\x40', '\x00\x41', '\x00\x42',
        '\x00\x43', '\x00\x44', '\x00\x45', '\x00\x46',
        '\x00\x60', '\x00\x61', '\x00\x62', '\x00\x63',
        '\x00\x64', '\x00\x65', '\x00\x66', '\x00\x67',
        '\x00\x68', '\x00\x69', '\x00\x6A', '\x00\x6B',
        '\x00\x6C', '\x00\x6D', '\x00\x80', '\x00\x81',
        '\x00\x82', '\x00\x83', '\x00\x84', '\x00\x85',
        '\x00\x86', '\x00\x87', '\x00\x88', '\x00\x89',
        '\x00\x8A', '\x00\x8B', '\x00\x8C', '\x00\x8D',
        '\x00\x8E', '\x00\x8F', '\x00\x90', '\x00\x91',
        '\x00\x92', '\x00\x93', '\x00\x94', '\x00\x95',
        '\x00\x96', '\x00\x97', '\x00\x98', '\x00\x99',
        '\x00\x9A', '\x00\x9B', '\x00\x9C', '\x00\x9D',
        '\x00\x9E', '\x00\x9F', '\x00\xA0', '\x00\xA1',
        '\x00\xA2', '\x00\xA3', '\x00\xA4', '\x00\xA5',
        '\x00\xA6', '\x00\xA7', '\x00\xA8', '\x00\xA9',
        '\x00\xAA', '\x00\xAB', '\x00\xAC', '\x00\xAD',
        '\x00\xAE', '\x00\xAF', '\x00\xB0', '\x00\xB1',
        '\x00\xB2', '\x00\xB3', '\x00\xB4', '\x00\xB5',
        '\x00\xB6', '\x00\xB7', '\x00\xB8', '\x00\xB9',
        '\x00\xBA', '\x00\xBB', '\x00\xBC', '\x00\xBD',
        '\x00\xBE', '\x00\xBF', '\x00\xC0', '\x00\xC1',
        '\x00\xC2', '\x00\xC3', '\x00\xC4', '\x00\xC5',
        '\x00\x00', '\xc0\x01', '\xc0\x02', '\xc0\x03',
        '\xc0\x04', '\xc0\x05', '\xc0\x06', '\xc0\x07',
        '\xc0\x08', '\xc0\x09', '\xc0\x0a', '\xc0\x0b',
        '\xc0\x0c', '\xc0\x0d', '\xc0\x0e', '\xc0\x0f',
        '\xc0\x10', '\xc0\x11', '\xc0\x12', '\xc0\x13',
        '\xc0\x14', '\xc0\x15', '\xc0\x16', '\xc0\x17',
        '\xc0\x18', '\xc0\x19', '\xC0\x1A', '\xC0\x1B',
        '\xC0\x1C', '\xC0\x1D', '\xC0\x1E', '\xC0\x1F',
        '\xC0\x20', '\xC0\x21', '\xC0\x22', '\xC0\x23',
        '\xC0\x24', '\xC0\x25', '\xC0\x26', '\xC0\x27',
        '\xC0\x28', '\xC0\x29', '\xC0\x2A', '\xC0\x2B',
        '\xC0\x2C', '\xC0\x2D', '\xC0\x2E', '\xC0\x2F',
        '\xC0\x30', '\xC0\x31', '\xC0\x32', '\xC0\x33',
        '\xC0\x34', '\xC0\x35', '\xC0\x36', '\xC0\x37',
        '\xC0\x38', '\xC0\x39', '\xC0\x3A', '\xC0\x3B',
        '\xfe\xfe', '\xfe\xff', '\xff\xe0', '\xff\xe1'
    ]

    # Create a TLS record out of a protocol packet
    def make_record(self, t, body):
        l = struct.pack("!H",len(body))
        return chr(t) + self.ssl_tokens[self._ssl_version] + l + body

    def make_hello(self):
        suites = "".join(self.ssl3_cipher)
        rand = "".join([ chr(int(256 * random.random())) for x in range(32) ])
        l  = struct.pack("!L", 39+len(suites))[1:] # 3 bytes
        sl = struct.pack("!H", len(suites))

        # Client hello, lenght and version
        # Random data + session ID + cipher suites + compression suites
        data  = "\x01" + l + self.ssl_tokens[self._ssl_version] + rand + "\x00"
        data += sl + suites + "\x01\x00"

        return self.make_record(22, data)

    def make_ccs(self):
        ccsbody = "\x01" # Empty CCS
        return self.make_record(20, ccsbody)

    def parse_handshake_pkt(self, buf):
        r = []
        while len(buf) >= 4:
            mt = ord(buf[0])
            mlen = struct.unpack("!L", buf[0:4])[0] & 0xFFFFFF

            if mlen+4 > len(buf):
                break

            r.append( {"type":mt, "data": buf[4:4+mlen]} )
            buf = buf[4+mlen:]
        return r

    def parse_alert_pkt(self, buf):
        return [ {"level": ord(buf[0]), "desc": ord(buf[1]) } ]

    def parse_records(self):
        r = []
        # 5 byte header
        while len(self._inbuffer) >= 5:
            mtype = ord(self._inbuffer[0])
            mtlsv = self._inbuffer[1:3]
            mlen  = struct.unpack("!H", self._inbuffer[3:5])[0]

            if len(self._inbuffer) < 5 + mlen:
                break

            if mtype == 22: # Handshake
                protp = self.parse_handshake_pkt(self._inbuffer[5:5 + mlen])
            elif mtype == 21: # Alert
                protp = self.parse_alert_pkt(self._inbuffer[5:5 + mlen])
            else:
                protp = []

            r.append( {"type":mtype, "sslv":mtlsv, "proto":protp} )

            self._inbuffer = self._inbuffer[5+mlen:]

        return r


class OpenSslCcsInjectionResult(PluginResult):
    """The result of running --openssl_ccs on a specific server.

    Attributes:
        is_vulnerable_to_ccs_injection (bool): True if the server is vulnerable to OpenSSL's CCS injection issue.
    """

    COMMAND_TITLE = 'OpenSSL CCS Injection'

    def __init__(self, server_info, plugin_command, plugin_options, is_vulnerable_to_ccs_injection):
        super(OpenSslCcsInjectionResult, self).__init__(server_info, plugin_command, plugin_options)
        self.is_vulnerable_to_ccs_injection = is_vulnerable_to_ccs_injection


    def as_xml(self):
        result_xml = Element(self.plugin_command, title=self.COMMAND_TITLE)
        result_xml.append(Element('openSslCcsInjection',
                                  attrib={'isVulnerable': str(self.is_vulnerable_to_ccs_injection)}))
        return result_xml


    def as_text(self):
        result_txt = [self.PLUGIN_TITLE_FORMAT(self.COMMAND_TITLE)]

        ccs_text = 'VULNERABLE - Server is vulnerable to OpenSSL CCS injection' \
            if self.is_vulnerable_to_ccs_injection \
            else 'OK - Not vulnerable to OpenSSL CCS injection'
        result_txt.append(self.FIELD_FORMAT('', ccs_text))
        return result_txt



# -*- coding: utf-8 -*-
"""Plugin to test the server for CVE-2014-0160.
"""



import new
from xml.etree.ElementTree import Element

from nassl import TLSV1, TLSV1_1, TLSV1_2, SSLV3
from nassl._nassl import WantX509LookupError, WantReadError

from sslyze.plugins import plugin_base
from sslyze.plugins.plugin_base import PluginResult
from sslyze.utils.ssl_connection import SSLHandshakeRejected


class HeartbleedPlugin(plugin_base.PluginBase):

    interface = plugin_base.PluginInterface("HeartbleedPlugin", "")
    interface.add_command(
        command="heartbleed",
        help="Tests the server(s) for the OpenSSL Heartbleed vulnerability (experimental)."
    )


    def process_task(self, server_info, command, options_dict=None):
        ssl_connection = server_info.get_preconfigured_ssl_connection()
        ssl_connection.ssl_version = server_info.highest_ssl_version_supported  # Needed by the heartbleed payload

        # Awful hack #1: replace nassl.sslClient.do_handshake() with a heartbleed
        # checking SSL handshake so that all the SSLyze options
        # (startTLS, proxy, etc.) still work
        ssl_connection.do_handshake = new.instancemethod(do_handshake_with_heartbleed, ssl_connection, None)

        heartbleed = None
        try: # Perform the SSL handshake
            ssl_connection.connect()
        except HeartbleedSent:
            # Awful hack #2: directly read the underlying network socket
            heartbleed = ssl_connection._sock.recv(16381)
        finally:
            ssl_connection.close()

        # Text output
        is_vulnerable_to_heartbleed = False
        if heartbleed is None:
            raise ValueError("Error: connection failed.")
        elif '\x01\x01\x01\x01\x01\x01\x01\x01\x01' in heartbleed:
            # Server replied with our hearbeat payload
            is_vulnerable_to_heartbleed = True

        return HeartbleedResult(server_info, command, options_dict, is_vulnerable_to_heartbleed)


class HeartbleedResult(PluginResult):
    """The result of running --heartbleed on a specific server.

    Attributes:
        is_vulnerable_to_heartbleed (bool): True if the server is vulnerable to the Heartbleed attack.
    """

    COMMAND_TITLE = 'OpenSSL Heartbleed'

    def __init__(self, server_info, plugin_command, plugin_options, is_vulnerable_to_heartbleed):
        super(HeartbleedResult, self).__init__(server_info, plugin_command, plugin_options)
        self.is_vulnerable_to_heartbleed = is_vulnerable_to_heartbleed

    def as_text(self):
        heartbleed_txt = 'VULNERABLE - Server is vulnerable to Heartbleed' \
            if self.is_vulnerable_to_heartbleed \
            else 'OK - Not vulnerable to Heartbleed'

        txt_output = [self.PLUGIN_TITLE_FORMAT(self.COMMAND_TITLE)]
        txt_output.append(self.FIELD_FORMAT("", heartbleed_txt))
        return txt_output

    def as_xml(self):
        xml_output = Element(self.plugin_command, title=self.COMMAND_TITLE)
        xml_output.append(Element('openSslHeartbleed', isVulnerable=str(self.is_vulnerable_to_heartbleed)))
        return xml_output



def heartbleed_payload(ssl_version):
    # This heartbleed payload does not exploit the server
    # https://blog.mozilla.org/security/2014/04/12/testing-for-heartbleed-vulnerability-without-exploiting-the-server/

    SSL_VERSION_MAPPING = {
        SSLV3 :  '\x00', # Surprising that it works with SSL 3 which doesn't define TLS extensions
        TLSV1 :  '\x01',
        TLSV1_1: '\x02',
        TLSV1_2: '\x03'}

    payload = (
        '\x18'           # Record type - Heartbeat
        '\x03{0}'               # TLS version
        '\x40\x00'              # Record length
        '\x01'                  # Heartbeat type - Request
        '\x3f\xfd')             # Heartbeat length

    payload += '\x01'*16381     # Heartbeat data

    payload += (                # Second Heartbeat request with no padding
        '\x18'                  # Record type - Heartbeat
        '\x03{0}'
        '\x00\x03\x01\x00\x00'
    )

    return payload.format(SSL_VERSION_MAPPING[ssl_version])


class HeartbleedSent(SSLHandshakeRejected):
    # Awful hack #3: Use an exception to hack the handshake's control flow in
    # a super obscure way
    pass


def do_handshake_with_heartbleed(self):
    # This is nassl's code for do_handshake() modified to send a heartbleed
    # payload that will send the heartbleed checking payload
    # I copied nassl's code here so I could leave anything heartbleed-related
    # outside of the nassl code base
    try:
        if self._ssl.do_handshake() == 1:
            self._handshakeDone = True
            return True # Handshake was successful

    except WantReadError:
        # OpenSSL is expecting more data from the peer
        # Send available handshake data to the peer
        # In this heartbleed handshake we only send the client hello
        lenToRead = self._network_bio.pending()
        while lenToRead:
            # Get the data from the SSL engine
            handshakeDataOut = self._network_bio.read(lenToRead)
            # Send it to the peer
            self._sock.send(handshakeDataOut)
            lenToRead = self._network_bio.pending()

        # Send the heartbleed payload after the client hello
        self._sock.send(heartbleed_payload(self.ssl_version))

        # Recover the peer's encrypted response
        # In this heartbleed handshake we only receive the server hello
        handshakeDataIn = self._sock.recv(2048)
        if len(handshakeDataIn) == 0:
            raise IOError('Nassl SSL handshake failed: peer did not send data back.')
        # Pass the data to the SSL engine
        self._network_bio.write(handshakeDataIn)

        # Signal that we sent the heartbleed payload and just stop the handshake
        raise HeartbleedSent("")


    except WantX509LookupError:
        # Server asked for a client certificate and we didn't provide one
        # Heartbleed should work anyway
        self._sock.send(heartbleed_payload(self.ssl_version)) # The heartbleed payload
        raise HeartbleedSent("") # Signal that we sent the heartbleed payload




# -*- coding: utf-8 -*-
"""Plugin to analyze the server's SSL session resumption capabilities.
"""

from xml.etree.ElementTree import Element

from nassl import SSL_OP_NO_TICKET

from sslyze.plugins import plugin_base
from sslyze.plugins.plugin_base import PluginResult
from sslyze.utils.thread_pool import ThreadPool


class SessionResumptionPlugin(plugin_base.PluginBase):

    interface = plugin_base.PluginInterface(
        title="SessionResumptionPlugin",
        description="Analyzes the target server's SSL session resumption capabilities."
    )
    interface.add_command(
        command="resum",
        help="Tests the server(s) for session resumption support using session IDs and TLS session tickets (RFC 5077)."
    )
    interface.add_command(
        command="resum_rate",
        help="Performs 100 session resumptions with the server(s), in order to estimate the session resumption rate.",
        aggressive=True
    )


    MAX_THREADS_NB = 20

    def process_task(self, server_info, command, options_dict=None):

        if command == 'resum':
            result = self._command_resum(server_info)
        elif command == 'resum_rate':
            successful_resumptions_nb, errored_resumptions_list = self._test_session_resumption_rate(server_info, 100)
            result = ResumptionRateResult(server_info, command, options_dict, 100, successful_resumptions_nb,
                                          errored_resumptions_list)
        else:
            raise ValueError("PluginSessionResumption: Unknown command.")

        return result


    def _command_resum(self, server_info):
        """Tests the server for session resumption support using session IDs and TLS session tickets (RFC 5077).
        """
        # Test Session ID support
        successful_resumptions_nb, errored_resumptions_list = self._test_session_resumption_rate(server_info, 5)

        # Test TLS tickets support
        ticket_exception = None
        ticket_reason = None
        ticket_supported = False
        try:
            (ticket_supported, ticket_reason) = self._resume_with_session_ticket(server_info)
        except Exception as e:
            ticket_exception = e

        return ResumptionResult(server_info, 'resum', {}, 5,  successful_resumptions_nb, errored_resumptions_list,
                                ticket_supported, ticket_reason, ticket_exception)


    def _test_session_resumption_rate(self, server_info, resumption_attempts_nb):
        """Attempts several session ID resumption with the server."""
        thread_pool = ThreadPool()

        for _ in xrange(resumption_attempts_nb):
            thread_pool.add_job((self._resume_with_session_id, (server_info, )))
        thread_pool.start(nb_threads=min(resumption_attempts_nb, self.MAX_THREADS_NB))

        # Count successful/failed resumptions
        successful_resumptions_nb = 0
        for completed_job in thread_pool.get_result():
            (job, was_resumption_successful) = completed_job
            if was_resumption_successful:
                successful_resumptions_nb += 1

        # Count errors and store error messages
        errored_resumptions_list = []
        for failed_job in thread_pool.get_error():
            (job, exception) = failed_job
            error_msg = '{} - {}'.format(str(exception.__class__.__name__), str(exception))
            errored_resumptions_list.append(error_msg)

        thread_pool.join()
        return successful_resumptions_nb, errored_resumptions_list


    def _resume_with_session_id(self, server_info):
        """Performs one session resumption using Session IDs.
        """
        session1 = self._resume_ssl_session(server_info)
        try:
            # Recover the session ID
            session1_id = self._extract_session_id(session1)
        except IndexError:
            # Session ID not assigned
            return False

        if session1_id == '':
            # Session ID empty
            return False

        # Try to resume that SSL session
        session2 = self._resume_ssl_session(server_info, session1)
        try: # Recover the session ID
            session2_id = self._extract_session_id(session2)
        except IndexError:
            # Session ID not assigned
            return False

        # Finally, compare the two Session IDs
        if session1_id != session2_id:
            # Session ID assigned but not accepted
            return False

        return True


    def _resume_with_session_ticket(self, server_info):
        """Performs one session resumption using TLS Session Tickets.
        """
        # Connect to the server and keep the SSL session
        session1 = self._resume_ssl_session(server_info, should_enable_tls_ticket=True)
        try:
            # Recover the TLS ticket
            session1_tls_ticket = self._extract_tls_session_ticket(session1)
        except IndexError:
            return False, 'TLS ticket not assigned'

        # Try to resume that session using the TLS ticket
        session2 = self._resume_ssl_session(server_info, session1, should_enable_tls_ticket=True)
        try:
            # Recover the TLS ticket
            session2_tls_ticket = self._extract_tls_session_ticket(session2)
        except IndexError:
            return False, 'TLS ticket not assigned'

        # Finally, compare the two TLS Tickets
        if session1_tls_ticket != session2_tls_ticket:
            return False, 'TLS ticket assigned but not accepted'

        return True, ''


    @staticmethod
    def _extract_session_id(ssl_session):
        """Extracts the SSL session ID from a SSL session object or raises IndexError if the session ID was not set.
        """
        session_string = ( (ssl_session.as_text()).split("Session-ID:") )[1]
        session_id = ( session_string.split("Session-ID-ctx:") )[0].strip()
        return session_id


    @staticmethod
    def _extract_tls_session_ticket(ssl_session):
        """Extracts the TLS session ticket from a SSL session object or raises IndexError if the ticket was not set.
        """
        session_string = ((ssl_session.as_text()).split("TLS session ticket:"))[1]
        session_tls_ticket = (session_string.split("Compression:"))[0]
        return session_tls_ticket


    def _resume_ssl_session(self, server_info, ssl_session=None, should_enable_tls_ticket=False):
        """Connects to the server and returns the session object that was assigned for that connection.
        If ssl_session is given, tries to resume that session.
        """
        ssl_connection = server_info.get_preconfigured_ssl_connection()
        if not should_enable_tls_ticket:
        # Need to disable TLS tickets to test session IDs, according to rfc5077:
        # If a ticket is presented by the client, the server MUST NOT attempt
        # to use the Session ID in the ClientHello for stateful session resumption
            ssl_connection.set_options(SSL_OP_NO_TICKET) # Turning off TLS tickets.

        if ssl_session:
            ssl_connection.set_session(ssl_session)

        try: # Perform the SSL handshake
            ssl_connection.connect()
            new_session = ssl_connection.get_session() # Get session data
        finally:
            ssl_connection.close()

        return new_session


class ResumptionRateResult(PluginResult):
    """The result of running --resum_rate on a specific server.

    Attributes:
        attempted_resumptions_nb (int): The total number of session ID resumptions that were attempted.
        successful_resumptions_nb (int): The number of session ID resumptions that were successful.
        failed_resumptions_nb (int): The number of session ID resumptions that failed.
        errored_resumptions_list (Optional[List[(str)]): A list of unexpected errors triggered while trying to perform
        session ID resumption with the server (should always be empty).
    """

    COMMAND_TITLE = 'Resumption Rate'

    def __init__(self, server_info, plugin_command, plugin_options, attempted_resumptions_nb, successful_resumptions_nb,
                 errored_resumptions_list):
        super(ResumptionRateResult, self).__init__(server_info, plugin_command, plugin_options)

        self.attempted_resumptions_nb = attempted_resumptions_nb
        self.successful_resumptions_nb = successful_resumptions_nb
        self.errored_resumptions_list = errored_resumptions_list
        self.failed_resumptions_nb = attempted_resumptions_nb - successful_resumptions_nb - \
                                     len(errored_resumptions_list)


    RESUMPTION_RESULT_FORMAT = '{4} ({0} successful, {1} failed, {2} errors, {3} total attempts).'.format
    RESUMPTION_LINE_FORMAT = '      {resumption_type:<35}{result}'.format
    RESUMPTION_ERROR_FORMAT = '        ERROR #{error_nb}: {error_msg}'.format

    def as_text(self):
        result_txt = [self.PLUGIN_TITLE_FORMAT(self.COMMAND_TITLE)]

        # Create the line which summarizes the session resumption rate
        if self.successful_resumptions_nb == self.attempted_resumptions_nb:
            resumption_supported_txt = 'OK - Supported'
        elif self.successful_resumptions_nb > 0:
            resumption_supported_txt = 'PARTIALLY SUPPORTED'
        elif self.failed_resumptions_nb == self.attempted_resumptions_nb:
            resumption_supported_txt = 'NOT SUPPORTED'
        else:
            resumption_supported_txt = 'ERROR'

        resum_rate_txt = self.RESUMPTION_RESULT_FORMAT(str(self.successful_resumptions_nb),
                                                      str(self.failed_resumptions_nb),
                                                      str(len(self.errored_resumptions_list)),
                                                      str(self.attempted_resumptions_nb),
                                                      resumption_supported_txt)
        result_txt.append(self.RESUMPTION_LINE_FORMAT(resumption_type='With Session IDs:', result=resum_rate_txt))

        # Add error messages if there was any
        i = 0
        for error_msg in self.errored_resumptions_list:
            result_txt.append(self.RESUMPTION_ERROR_FORMAT(error_nb=i, error_msg=error_msg))
            i += 1

        return result_txt


    def as_xml(self):
        xml_result = Element(self.plugin_command, title=self.COMMAND_TITLE)

        resumption_rate_xml = Element(
                'sessionResumptionWithSessionIDs',
                attrib={'totalAttempts': str(self.attempted_resumptions_nb),
                        'errors': str(len(self.errored_resumptions_list)),
                        'isSupported': str(self.attempted_resumptions_nb == self.successful_resumptions_nb),
                        'successfulAttempts': str(self.successful_resumptions_nb),
                        'failedAttempts': str(self.failed_resumptions_nb)}
        )
        # Add error messages if there was any
        for error_msg in self.errored_resumptions_list:
            resumption_error_xml = Element('error')
            resumption_error_xml.text = error_msg
            resumption_rate_xml.append(resumption_error_xml)

        xml_result.append(resumption_rate_xml)
        return xml_result



class ResumptionResult(ResumptionRateResult):
    """The result of running --resum on a specific server; also has all the attributes of ResumptionRateResult.

    Attributes:
        is_ticket_resumption_supported (bool): True if the server honors client-initiated renegotiation attempts.
        ticket_resumption_failed_reason (str): A message explaining why TLS ticket resumption failed.
        ticket_resumption_exception (Optional[str]): An unexpected error that was raised while trying to perform ticket
            resumption (should never happen).
    """

    def __init__(self, server_info, plugin_command, plugin_options, attempted_resumptions_nb, successful_resumptions_nb,
                 errored_resumptions_list, is_ticket_resumption_supported, ticket_resumption_failed_reason=None,
                 ticket_resumption_exception=None):

        super(ResumptionResult, self).__init__(server_info, plugin_command, plugin_options,
                                               attempted_resumptions_nb, successful_resumptions_nb,
                                               errored_resumptions_list)

        self.is_ticket_resumption_supported = is_ticket_resumption_supported
        self.ticket_resumption_failed_reason = ticket_resumption_failed_reason

        # An exception was raised while trying to perform ticket resumption (should never happen)
        self.ticket_resumption_error = None
        if ticket_resumption_exception:
            self.ticket_resumption_error = '{} - {}'.format(str(ticket_resumption_exception.__class__.__name__),
                                                            str(ticket_resumption_exception))

    COMMAND_TITLE = 'Session Resumption'

    def as_text(self):
        # Same output as --resum_rate but add a line about TLS ticket resumption at the end
        result_txt = super(ResumptionResult, self).as_text()

        if self.ticket_resumption_error:
            ticket_txt = 'ERROR: {}'.format(self.ticket_resumption_error)
        else:
            ticket_txt = 'OK - Supported' \
                if self.is_ticket_resumption_supported \
                else 'NOT SUPPORTED - {}.'.format(self.ticket_resumption_failed_reason)


        result_txt.append(self.RESUMPTION_LINE_FORMAT(resumption_type='With TLS Tickets:', result=ticket_txt))
        return result_txt


    def as_xml(self):
        xml_result = Element(self.plugin_command, title=self.COMMAND_TITLE)

        # We keep the session resumption XML node
        resum_rate_xml = super(ResumptionResult, self).as_xml()
        session_resum_xml = resum_rate_xml[0]
        xml_result.append(session_resum_xml)

        # Add the ticket resumption node
        xml_resum_ticket_attr = {}
        if self.ticket_resumption_error:
            xml_resum_ticket_attr['error'] = self.ticket_resumption_error
        else:
            xml_resum_ticket_attr['isSupported'] = str(self.is_ticket_resumption_supported)
            if not self.is_ticket_resumption_supported:
                xml_resum_ticket_attr['reason'] = self.ticket_resumption_failed_reason

        xml_resum_ticket = Element('sessionResumptionWithTLSTickets', attrib=xml_resum_ticket_attr)
        xml_result.append(xml_resum_ticket)

        return xml_result


# -*- coding: utf-8 -*-
"""Utility class to discover the list of available plugins.
"""

import importlib
import inspect
import sys
import sslyze.plugins
import sslyze.plugins.plugin_base


class PluginsFinder:

    def __init__(self):
        """Finds available plugins by discovering any class that implements the PluginBase abstract class.

        Returns PluginsFinder: An object encapsulating the list of available sslyze plugin classess.
        """
        self._plugin_classes = set([])
        self._commands = {}
        self._aggressive_comands = []


        if hasattr(sys,"frozen") and sys.frozen in ("windows_exe", "console_exe"):
            # For py2exe builds we have to load the plugins statically using a hardcoded list
            plugin_modules = self.get_plugin_modules_static()
        else:
            # When ran from the interpreter, just dynamically find the available plugins
            plugin_modules = self.get_plugin_modules_dynamic()

        for module in plugin_modules:
            # Check every declaration in that module
            for name in dir(module):
                obj = getattr(module, name)

                if inspect.isclass(obj):
                    # A class declaration was found in that module; checking if it's a subclass of PluginBase
                    # Discarding PluginBase as a subclass of PluginBase
                    if obj != sslyze.plugins.plugin_base.PluginBase:
                        for base in obj.__bases__:
                            # H4ck because issubclass() doesn't seem to work as expected on Linux
                            # It has to do with PluginBase being imported multiple times (within plugins) or something
                            if base.__name__ == 'PluginBase':
                                # A plugin was found, keep it
                                self._plugin_classes.add(obj)

                        #if issubclass(obj, plugins.PluginBase.PluginBase):
                            # A plugin was found, keep it
                        #    self._plugin_classes.add(obj)

                                # Store the plugin's commands
                                for (cmd, is_aggressive) in obj.get_interface().get_commands_as_text():
                                    self._commands[cmd] = obj
                                    # Store a list of aggressive commands
                                    if is_aggressive:
                                        self._aggressive_comands.append(cmd)



    def get_plugins(self):
        return self._plugin_classes


    def get_commands(self):
        return self._commands


    def get_aggressive_commands(self):
        return self._aggressive_comands


    @staticmethod
    def get_plugin_modules_static():

        import sslyze.plugins.certificate_info_plugin
        import sslyze.plugins.compression_plugin
        import sslyze.plugins.fallback_scsv_plugin
        import sslyze.plugins.heartbleed_plugin
        import sslyze.plugins.hsts_plugin
        import sslyze.plugins.openssl_ccs_injection_plugin
        import sslyze.plugins.openssl_cipher_suites_plugin
        import sslyze.plugins.session_renegotiation_plugin
        import sslyze.plugins.session_resumption_plugin

        plugin_modules = [
            sslyze.plugins.certificate_info_plugin,
            sslyze.plugins.compression_plugin,
            sslyze.plugins.fallback_scsv_plugin,
            sslyze.plugins.heartbleed_plugin,
            sslyze.plugins.hsts_plugin,
            sslyze.plugins.openssl_ccs_injection_plugin,
            sslyze.plugins.openssl_cipher_suites_plugin,
            sslyze.plugins.session_renegotiation_plugin,
            sslyze.plugins.session_resumption_plugin
        ]

        return plugin_modules


    @staticmethod
    def get_plugin_modules_dynamic():
        plugin_modules = []
        import pkgutil
        for loader, module_name, _ in list(pkgutil.iter_modules(sslyze.plugins.__path__, prefix='sslyze.plugins.')):
            try:
                module = loader.find_module(module_name).load_module(module_name)
                plugin_modules.append(module)
            except Exception as e:
                print '  {module} - Import Error: {error}'.format(module=module_name, error=str(e))
                continue

        return plugin_modules



# -*- coding: utf-8 -*-
"""Main abstract plugin classes from which all the plugins should inherit.
"""

import abc
from optparse import make_option
from xml.etree.ElementTree import Element


class PluginInterface(object):
    """This object tells SSLyze what the plugin does: its title, description, and which command line option(s) it
    implements.

    Every plugin should have a class attribute called interface that is an instance of PluginInterface.
    """

    def __init__(self, title, description):
        """
        Title and description are sent to optparse.OptionGroup().
        """
        self.title = title
        self.description = description
        self._options = []
        self._commands = []
        self._commands_as_text = []

    def add_option(self, option, help, dest=None):
        """Options are settings specific to one single plugin; they will passed to process_task() in the options_dict.
        """
        # If dest is something, store it, otherwise just use store_true
        action="store_true"
        if dest is not None:
            action="store"

        self._options.append(make_option('--{}'.format(option), action=action, help=help, dest=dest))


    def add_command(self, command, help, aggressive=False):
        """Commands are actions/scans the plugin implements, with PluginXXX.process_task().

        Setting aggressive to True is a warning that the command will open many simultaneous connections to the server
        and should therefore not be run concurrently with other `aggressive` commands against a given server.
        """

        self._commands.append(make_option('--{}'.format(command), action='store_true', help=help))
        self._commands_as_text.append((command, aggressive))

    @staticmethod
    def _make_option(command, help, dest):
        # If dest is something, store it, otherwise just use store_true
        action="store_true"
        if dest is not None:
            action="store"

        return make_option('--' + command, action=action, help=help, dest=dest)


    def get_commands(self):
        return self._commands


    def get_commands_as_text(self):
        return self._commands_as_text


    def get_options(self):
        return self._options


class PluginResult(object):
    """Plugins should return the result of process_task() as a subclass of this.
    """
    __metaclass__ = abc.ABCMeta

    # Common formatting
    PLUGIN_TITLE_FORMAT = '  * {0}:'.format
    FIELD_FORMAT = '      {0:<35}{1}'.format

    def __init__(self, server_info, plugin_command, plugin_options):

        self.server_info = server_info
        self.plugin_command = plugin_command
        self.plugin_options = plugin_options

    @abc.abstractmethod
    def as_xml(self):
        return

    @abc.abstractmethod
    def as_text(self):
        return


class PluginRaisedExceptionResult(PluginResult):
    """Returned when a plugin threw an exception while doing process_task()."""

    def __init__(self, server_info, plugin_command, plugin_options, exception):
        super(PluginRaisedExceptionResult, self).__init__(server_info, plugin_command, plugin_options)
        # Cannot keep the full exception as it may not be pickable (ie. _nassl.OpenSSLError)
        self.error_message = '{} - {}'.format(str(exception.__class__.__name__), str(exception))

    TITLE_TXT_FORMAT = 'Unhandled exception while running --{command}:'.format

    def as_text(self):
        return [self.TITLE_TXT_FORMAT(command=self.plugin_command), self.error_message]

    def as_xml(self):
        return Element(self.plugin_command, exception=self.as_text()[1])


class PluginBase(object):
    """Base plugin abstract class. All plugins have to inherit from it.
    """
    __metaclass__ = abc.ABCMeta

    # To be defined in subclasses
    interface = None

    @classmethod
    def get_interface(plugin_class):
        """Returns the AvailableCommands object for the current plugin.
        """
        return plugin_class.interface

    @abc.abstractmethod
    def process_task(self, server_connectivity_info, command, options_dict=None):
        """Completes the task specified by command on the server.

        Args:
            server_connectivity_info (ServerConnectivityInfo):
            command (str):
            options_dict (dict):
        """
        return



# -*- coding: utf-8 -*-
"""The main process class responsible for instantiating and running the plugins.
"""


from multiprocessing import Process
from sslyze.utils.ssl_connection import SSLConnection


class WorkerProcess(Process):

    def __init__(self, priority_queue_in, queue_in, queue_out, available_commands, network_retries, network_timeout):
        Process.__init__(self)
        self.priority_queue_in = priority_queue_in
        self.queue_in = queue_in
        self.queue_out = queue_out
        self.available_commands = available_commands

        # Set global network settings; needs to be done in each process
        SSLConnection.set_global_network_settings(network_retries, network_timeout)

    def run(self):
        """The process will first complete tasks it gets from self.queue_in.
        Once it gets notified that all the tasks have been completed, it terminates.
        """
        from sslyze.plugins.plugin_base import PluginRaisedExceptionResult

        # Start processing task in the priority queue first
        current_queue_in = self.priority_queue_in
        while True:

            task = current_queue_in.get()  # Grab a task from queue_in
            if task is None:  # All tasks have been completed
                current_queue_in.task_done()

                if current_queue_in == self.priority_queue_in:
                    # All high priority tasks have been completed; switch to low priority tasks
                    current_queue_in = self.queue_in
                    continue
                else:
                    # All the tasks have been completed; pass on the sentinel to result_queue and exit
                    self.queue_out.put(None)
                    break

            server_info, command, options_dict = task
            # Instantiate the proper plugin
            plugin_instance = self.available_commands[command]()

            try:
                # Process the task
                result = plugin_instance.process_task(server_info, command, options_dict)
            except Exception as e:
                #raise
                result = PluginRaisedExceptionResult(server_info, command, options_dict, e)

            # Send the result to queue_out
            self.queue_out.put(result)
            current_queue_in.task_done()

        return


#!/usr/bin/env python2.7
from sslyze import PROJECT_URL, PROJECT_DESC, __author__, __email__, __version__, __license__


SSLYZE_SETUP = {
    'name': 'SSLyze',
    'version': __version__,
    'description': PROJECT_DESC,
    'author': __author__,
    'author_email': __email__,
    'license': __license__,
    'url': PROJECT_URL,
    'scripts': ['sslyze_cli.py'],
    'packages': ['sslyze', 'sslyze.plugins', 'sslyze.utils'],
    'package_data': {'sslyze.plugins': ['data/trust_stores/*.pem']},
    'install_requires': ['nassl>=0.13.0,<1.14.0'],
    'classifiers': [
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'Natural Language :: French',
        'License :: OSI Approved :: GNU General Public License v2 (GPLv2)',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Topic :: System :: Networking',
        'Topic :: Security'
    ],
}

if __name__ == "__main__":
    # Importing setuptools here because setup_py2exe also imports SSLYZE_SETUP but needs to use distutils
    from setuptools import setup

    setup(**SSLYZE_SETUP)



# -*- coding: utf-8 -*-
"""Settings to be used for configuring an SSL connection via the ServerConnectivityInfo class.
"""

import os
from base64 import b64encode
from urllib import quote
from urlparse import urlparse

from nassl import SSL_FILETYPE_PEM, SSL_FILETYPE_ASN1
from nassl.ssl_client import SslClient


class TlsWrappedProtocolEnum(object):
    """The list of TLS-wrapped protocols supported by SSLyze; used to figure out how to establish an SSL connection to
    the server and what kind of "hello" message to send after the handshake was completed.
    """
    PLAIN_TLS = 1  # Standard TLS connection
    HTTPS = 2
    STARTTLS_SMTP = 3
    STARTTLS_XMPP = 4
    STARTTLS_XMPP_SERVER = 5
    STARTTLS_FTP = 6
    STARTTLS_POP3 = 7
    STARTTLS_LDAP = 8
    STARTTLS_IMAP = 9
    STARTTLS_RDP = 10
    STARTTLS_POSTGRES = 11


class ClientAuthenticationCredentials(object):
    """Everything needed to perform client authentication with an SSL server.
    """

    def __init__(self, client_certificate_chain_path, client_key_path, client_key_type=SSL_FILETYPE_PEM,
                 client_key_password=''):

        self.client_certificate_chain_path = client_certificate_chain_path
        if not os.path.isfile(self.client_certificate_chain_path):
            raise ValueError('Could not open the client certificate file')

        self.client_key_path = client_key_path
        if not os.path.isfile(self.client_key_path):
            raise ValueError('Could not open the client private key file')

        self.client_key_password = client_key_password

        self.client_key_type = client_key_type
        if self.client_key_type not in [SSL_FILETYPE_ASN1, SSL_FILETYPE_PEM]:
            raise ValueError('Invalid certificate format specified')

        # Try to load the cert and key in OpenSSL; will raise an exception if something is wrong
        SslClient(client_certchain_file=self.client_certificate_chain_path, client_key_file=self.client_key_path,
                  client_key_type=self.client_key_type, client_key_password=self.client_key_password)


class HttpConnectTunnelingSettings(object):
    """Information needed to tunnel SSL/TLS traffic through an HTTP Connect Proxy.
    """

    def __init__(self, hostname, port, basic_auth_user=None, basic_auth_password=None):
        self.hostname = hostname
        self.port = port
        self.basic_auth_user = basic_auth_user
        self.basic_auth_password = basic_auth_password


    @classmethod
    def from_url(cls, proxy_url):
        parsed_url = urlparse(proxy_url)

        if not parsed_url.netloc or not parsed_url.hostname:
            raise ValueError('Invalid Proxy URL.')

        if parsed_url.scheme == 'http':
           default_port = 80
        elif parsed_url.scheme == 'https':
           default_port = 443
        else:
            raise ValueError('Invalid URL scheme')

        port = parsed_url.port if parsed_url.port else default_port
        return cls(parsed_url.hostname, port, parsed_url.username, parsed_url.password)


    def get_basic_auth_header(self):
        """Generate the right HTTP header for Basic Authentication.
        """
        header = ''
        if self.basic_auth_user is not None:
            header = b64encode('{0}:{1}'.format(quote(self.basic_auth_user), quote(self.basic_auth_password)))
        return header